@@ -22,6 +22,7 @@ from mercurial import (
     commands,
     context,
     encoding,
+    phases,
     util as hgutil,
     url,
 )
@@ -104,18 +105,22 @@ class GitHandler(object):
     def __init__(self, dest_repo, ui):
         self.repo = dest_repo
         self.ui = ui
+        self.vfs = self.repo.vfs
 
-        if ui.configbool('git', 'intree'):
+        # Mercurial >= 3.3:  repo.shared()
+        if dest_repo.sharedpath != dest_repo.path:
+            self.vfs = compat.hgvfs(dest_repo.sharedpath)
+        if compat.config(ui, 'bool', 'git', 'intree'):
             self.gitdir = self.repo.wvfs.join('.git')
         else:
-            self.gitdir = self.repo.vfs.join('git')
+            self.gitdir = self.vfs.join('git')
 
         self.init_author_file()
 
         self.paths = ui.configitems('paths')
 
-        self.branch_bookmark_suffix = ui.config('git',
-                                                'branch_bookmark_suffix')
+        self.branch_bookmark_suffix = compat.config(
+            ui, 'string', 'git', 'branch_bookmark_suffix')
 
         self._map_git_real = None
         self._map_hg_real = None
@@ -158,8 +163,9 @@ class GitHandler(object):
 
     def init_author_file(self):
         self.author_map = {}
-        if self.ui.config('git', 'authors'):
-            with open(self.repo.wvfs.join(self.ui.config('git', 'authors'))) as f:
+        authors_path = compat.config(self.ui, 'string', 'git', 'authors')
+        if authors_path:
+            with open(self.repo.wvfs.join(authors_path)) as f:
                 for line in f:
                     line = line.strip()
                     if not line or line.startswith('#'):
@@ -182,8 +188,8 @@ class GitHandler(object):
     def load_map(self):
         map_git_real = {}
         map_hg_real = {}
-        if os.path.exists(self.repo.vfs.join(self.map_file)):
-            for line in self.repo.vfs(self.map_file):
+        if os.path.exists(self.vfs.join(self.map_file)):
+            for line in self.vfs(self.map_file):
                 # format is <40 hex digits> <40 hex digits>\n
                 if len(line) != 82:
                     raise ValueError(
@@ -199,7 +205,7 @@ class GitHandler(object):
     def save_map(self, map_file):
         wlock = self.repo.wlock()
         try:
-            file = self.repo.vfs(map_file, 'w+', atomictemp=True)
+            file = self.vfs(map_file, 'w+', atomictemp=True)
             map_hg = self._map_hg
             buf = cStringIO.StringIO()
             bwrite = buf.write
@@ -214,13 +220,13 @@ class GitHandler(object):
 
     def load_tags(self):
         self.tags = {}
-        if os.path.exists(self.repo.vfs.join(self.tags_file)):
-            for line in self.repo.vfs(self.tags_file):
+        if os.path.exists(self.vfs.join(self.tags_file)):
+            for line in self.vfs(self.tags_file):
                 sha, name = line.strip().split(' ', 1)
                 self.tags[name] = sha
 
     def save_tags(self):
-        file = self.repo.vfs(self.tags_file, 'w+', atomictemp=True)
+        file = self.vfs(self.tags_file, 'w+', atomictemp=True)
         for name, sha in sorted(self.tags.iteritems()):
             if not self.repo.tagtype(name) == 'global':
                 file.write("%s %s\n" % (sha, name))
@@ -262,7 +268,7 @@ class GitHandler(object):
         self.save_map(self.map_file)
 
     def fetch(self, remote, heads):
-        refs = self.fetch_pack(remote, heads)
+        result = self.fetch_pack(remote, heads)
         remote_name = self.remote_name(remote)
 
         # if remote returns a symref for HEAD, then let's store that
@@ -270,18 +276,19 @@ class GitHandler(object):
         rnode = None
         oldheads = self.repo.changelog.heads()
         imported = 0
-        if refs:
-            filteredrefs = self.filter_min_date(self.filter_refs(refs, heads))
+        if result.refs:
+            filteredrefs = self.filter_min_date(self.filter_refs(result.refs,
+                                                                 heads))
             imported = self.import_git_objects(remote_name, filteredrefs)
-            self.import_tags(refs)
-            self.update_hg_bookmarks(refs)
+            self.import_tags(result.refs)
+            self.update_hg_bookmarks(result.refs)
 
             try:
-                symref = refs['HEAD']
+                symref = result.symrefs['HEAD']
                 if symref.startswith('refs/heads'):
                     rhead = symref.replace('refs/heads/', '')
 
-                rnode = refs['refs/heads/%s' % rhead]
+                rnode = result.refs['refs/heads/%s' % rhead]
                 rnode = self._map_git[rnode]
                 rnode = self.repo[rnode].node()
             except KeyError:
@@ -290,10 +297,10 @@ class GitHandler(object):
                 rnode = None
 
             if remote_name:
-                self.update_remote_branches(remote_name, refs)
-            elif not self.paths:
+                self.update_remote_branches(remote_name, result.refs)
+            elif not self.git.refs.as_dict('refs/remotes/'):
                 # intial cloning
-                self.update_remote_branches('default', refs)
+                self.update_remote_branches('default', result.refs)
 
                 # "Activate" a tipmost bookmark.
                 bms = self.repo['tip'].bookmarks()
@@ -316,6 +323,29 @@ class GitHandler(object):
 
         self.save_map(self.map_file)
 
+        # also mark public any branches the user specified
+        blist = [self.repo[branch].node() for branch in
+                 self.ui.configlist('git', 'public')]
+        if rnode and self.ui.configbool('hggit', 'usephases'):
+            blist.append(rnode)
+
+        if blist:
+            lock = self.repo.lock()
+            try:
+                tr = self.repo.transaction("phase")
+                try:
+                    phases.advanceboundary(self.repo, tr, phases.public,
+                                           blist)
+                except TypeError:
+                    # hg < 3.2
+                    phases.advanceboundary(self.repo, phases.public,
+                                           blist)
+                tr.close()
+            finally:
+                if tr is not None:
+                    tr.release()
+                lock.release()
+
         if imported == 0:
             return 0
 
@@ -406,7 +436,7 @@ class GitHandler(object):
         return ret
 
     def clear(self):
-        mapfile = self.repo.vfs.join(self.map_file)
+        mapfile = self.vfs.join(self.map_file)
         if os.path.exists(self.gitdir):
             for root, dirs, files in os.walk(self.gitdir, topdown=False):
                 for name in files:
@@ -420,21 +450,21 @@ class GitHandler(object):
     # incoming support
     def getremotechanges(self, remote, revs):
         self.export_commits()
-        refs = self.fetch_pack(remote.path, revs)
+        result = self.fetch_pack(remote.path, revs)
 
         # refs contains all remote refs. Prune to only those requested.
         if revs:
             reqrefs = {}
             for rev in revs:
                 for n in ('refs/heads/' + rev, 'refs/tags/' + rev):
-                    if n in refs:
-                        reqrefs[n] = refs[n]
+                    if n in result.refs:
+                        reqrefs[n] = result.refs[n]
         else:
-            reqrefs = refs
+            reqrefs = result.refs
 
         commits = [bin(c) for c in self.get_git_incoming(reqrefs).commits]
 
-        b = overlayrepo(self, commits, refs)
+        b = overlayrepo(self, commits, result.refs)
 
         return (b, commits, lambda: None)
 
@@ -489,7 +519,8 @@ class GitHandler(object):
         exporter = hg2git.IncrementalChangesetExporter(
             self.repo, pctx, self.git.object_store, gitcommit)
 
-        mapsavefreq = self.ui.configint('hggit', 'mapsavefrequency', 0)
+        mapsavefreq = compat.config(self.ui, 'int', 'hggit',
+                                          'mapsavefrequency')
         for i, ctx in enumerate(export):
             self.ui.progress('exporting', i, total=total)
             self.export_hg_commit(ctx.node(), exporter)
@@ -618,8 +649,10 @@ class GitHandler(object):
 
         >>> from collections import namedtuple
         >>> from mercurial.ui import ui
-        >>> mockrepo = namedtuple('localrepo', ['vfs'])
+        >>> mockrepo = namedtuple('localrepo', ['vfs', 'sharedpath', 'path'])
         >>> mockrepo.vfs = ''
+        >>> mockrepo.sharedpath = ''
+        >>> mockrepo.path = ''
         >>> g = GitHandler(mockrepo, ui()).get_valid_git_username_email
         >>> g('John Doe')
         'John Doe'
@@ -694,8 +727,8 @@ class GitHandler(object):
         # HG EXTRA INFORMATION
 
         # test only -- do not document this!
-        extra_in_message = self.ui.configbool('git', 'debugextrainmessage',
-                                              False)
+        extra_in_message = compat.config(self.ui, 'bool', 'git',
+                                         'debugextrainmessage')
         extra_message = ''
         git_extra = []
         if ctx.branch() != 'default':
@@ -776,7 +809,8 @@ class GitHandler(object):
         else:
             self.ui.status(_("no changes found\n"))
 
-        mapsavefreq = self.ui.configint('hggit', 'mapsavefrequency', 0)
+        mapsavefreq = compat.config(self.ui, 'int', 'hggit',
+                                          'mapsavefrequency')
         for i, csha in enumerate(commits):
             self.ui.progress('importing', i, total=total, unit='commits')
             commit = commit_cache[csha]
@@ -951,16 +985,10 @@ class GitHandler(object):
                 if copied:
                     copied_path = copied[0]
 
-            try:
-                return context.memfilectx(self.repo, f, data,
-                                          islink='l' in e,
-                                          isexec='x' in e,
-                                          copied=copied_path)
-            except TypeError:
-                return context.memfilectx(f, data,
-                                          islink='l' in e,
-                                          isexec='x' in e,
-                                          copied=copied_path)
+            return compat.memfilectx(self.repo, memctx, f, data,
+                                     islink='l' in e,
+                                     isexec='x' in e,
+                                     copied=copied_path)
 
         p1, p2 = (nullid, nullid)
         octopus = False
@@ -1196,8 +1224,7 @@ class GitHandler(object):
             progress = GitProgress(self.ui)
             f = StringIO.StringIO()
 
-            # monkey patch dulwich's read_pkt_refs so that we can determine on
-            # clone which bookmark to activate
+            # only newer versions (0.18) of dulwich have symref support
             client.read_pkt_refs = compat.read_pkt_refs
             ret = localclient.fetch_pack(path, determine_wants, graphwalker,
                                          f.write, progress.progress)
@@ -1209,7 +1236,22 @@ class GitHandler(object):
             # For empty repos dulwich gives us None, but since later
             # we want to iterate over this, we really want an empty
             # iterable
-            return ret if ret else {}
+            if ret is None:
+                ret = {}
+
+            # for older dulwich, the return type was a dict, meaning we're
+            # running our monkey patch read_pkt_refs
+            if isinstance(ret, dict):
+                head = ret.get("HEAD")
+                symrefs = {}
+                # we want to raise a KeyError if HEAD is not a key
+                if head:
+                    symrefs["HEAD"] = head
+                # use FetchPackResult as upstream now does
+                ret = compat.FetchPackResult(ret,
+                                             symrefs,
+                                             client.default_user_agent_string())
+            return ret
         except (HangupException, GitProtocolError), e:
             raise hgutil.Abort(_("git remote error: ") + str(e))
 
@@ -1260,7 +1302,7 @@ class GitHandler(object):
         '''filter refs by minimum date
 
         This only works for refs that are available locally.'''
-        min_date = self.ui.config('git', 'mindate')
+        min_date = compat.config(self.ui, 'string', 'git', 'mindate')
         if min_date is None:
             return refs
 
@@ -1545,21 +1587,21 @@ class GitHandler(object):
     @hgutil.propertycache
     def _rename_detector(self):
         # disabled by default to avoid surprises
-        similarity = self.ui.configint('git', 'similarity', default=0)
+        similarity = compat.config(self.ui, 'int', 'git', 'similarity')
         if similarity < 0 or similarity > 100:
-            raise util.Abort(_('git.similarity must be between 0 and 100'))
+            raise hgutil.Abort(_('git.similarity must be between 0 and 100'))
         if similarity == 0:
             return None
 
         # default is borrowed from Git
-        max_files = self.ui.configint('git', 'renamelimit', default=400)
-        if similarity < 0:
-            raise util.Abort(_('git.renamelimit must be non-negative'))
+        max_files = compat.config(self.ui, 'int', 'git', 'renamelimit')
+        if max_files < 0:
+            raise hgutil.Abort(_('git.renamelimit must be non-negative'))
         if max_files == 0:
             max_files = None
 
-        find_copies_harder = self.ui.configbool('git', 'findcopiesharder',
-                                                default=False)
+        find_copies_harder = compat.config(self.ui, 'bool', 'git',
+                                           'findcopiesharder')
         return diff_tree.RenameDetector(self.git.object_store,
                                         rename_threshold=similarity,
                                         max_files=max_files,
@@ -1606,7 +1648,7 @@ class GitHandler(object):
 
     def audit_hg_path(self, path):
         if '.hg' in path.split(os.path.sep):
-            if self.ui.configbool('git', 'blockdothg', True):
+            if compat.config(self.ui, 'bool', 'git', 'blockdothg'):
                 raise hgutil.Abort(
                     ('Refusing to import problematic path %r' % path),
                     hint=("Mercurial cannot check out paths inside nested " +
@@ -1648,8 +1690,10 @@ class GitHandler(object):
         >>> from collections import namedtuple
         >>> from dulwich.client import HttpGitClient, SSHGitClient
         >>> from mercurial.ui import ui
-        >>> mockrepo = namedtuple('localrepo', ['vfs'])
+        >>> mockrepo = namedtuple('localrepo', ['vfs', 'sharedpath', 'path'])
         >>> mockrepo.vfs = ''
+        >>> mockrepo.sharedpath = ''
+        >>> mockrepo.path = ''
         >>> g = GitHandler(mockrepo, ui())
         >>> client, url = g.get_transport_and_path('http://fqdn.com/test.git')
         >>> print isinstance(client, HttpGitClient)
